// errorcheck

// Copyright 2020 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package p

import "unsafe"

type T struct { // ERROR "recursive type"
	x int
	p unsafe.Pointer

	f T
}
