// Code generated by go tool dist; DO NOT EDIT.

package buildcfg

import "runtime"

const defaultGO386 = `sse2`
const defaultGOAMD64 = `v1`
const defaultGOARM = `5`
const defaultGOMIPS = `hardfloat`
const defaultGOMIPS64 = `hardfloat`
const defaultGOPPC64 = `power8`
const defaultGOEXPERIMENT = ``
const defaultGO_EXTLINK_ENABLED = ``
const defaultGO_LDSO = `/lib64/ld-linux-x86-64.so.2`
const version = `go1.19.7`
const defaultGOOS = runtime.GOOS
const defaultGOARCH = runtime.GOARCH
