// rundir

// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Issue 1536: bug when handling imported interfaces with
// private methods.

package ignored
