// compile

// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package main

func main() {
	testRecover()

}

func testRecover() {
	if false {
		func() {
			defer func() {
				_ = recover()
			}()
		}()
	}
}
