// Copyright 2013 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

#include <pthread.h>

static __thread int tls;

void
setTLS(int v)
{
	tls = v;
}

int
getTLS()
{
	return tls;
}
