// Copyright 2014 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package a

const (
	sinPi4 = 0.70710678118654752440084436210484903928483593768847
	A = complex(sinPi4, -sinPi4)
)

