// +build ignore

// Copyright 2015 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// For linkmain_run.go.

package notmain

func main() {
}
